# Python implementation of the HC Fused algorithm

def get_mean():
    pass

def get_i_j():
    pass

def get_sample():
    pass

def hc_fused():
    pass

def which_is_3():
    pass

def which_vector_matrix():
    pass

def hc_fused_calculate_distances():
    pass
